defmodule Indian.Parser do
  
  @moduledoc """ 

  A HTTP/1.1 request parser, using the [NimbleParsec
  library](https://github.com/dashbitco/nimble_parsec). [Official
  documentation](https://hexdocs.pm/nimble_parsec/NimbleParsec.html). [One
  good tutorial on
  NImbleParsec](https://serokell.io/blog/parser-combinators-in-elixir)
  and [another
  one](https://blog.drewolson.org/parser-combinators-in-elixir). The
  reference for HTTP 1.1 is RFC 9112 but also RFC 9110.

  """
  
  import NimbleParsec

  method = ascii_string([not: ?\ ], min: 1)

  path = ascii_string([not: ?\ ], min: 1)

  # We insist on upper case (RFC 9112, section 2.3 and appendix A)
  version = ignore(string("HTTP/")) |> ascii_string([not: ?\r, not: ?\n, not: ?\ ], min: 1)
  
  eol = string("\r\n") # We are not liberal here (we could define eol
		       # as a choice between \r\n and \n) but RFC 9112
		       # is clear (section 2.1)

  header_name = ascii_string([?!, ?#, ?$, ?%, ?&, ?', ?*, ?+, ?-, ?.,
			      ?^, ?_, ?`, ?|, ?~,
			      ?a..?z, ?A..?Z, ?0..?9 ], min: 1) # RFC
             # 9110, section 5.1 and
             # appendix A.
  header_value = ascii_string([?\s, ?\t, ?!..?~, ? ..?ÿ, 0x80..0xFF], min: 0) # Field
              # values can be any visible ASCII character, spaces and
              # tabs, *and* old Latin-1 characters (in theory, even
              # starting at 0x80 but we started only at 0xA0). We
              # cannot use utf8_string, header values are bytes, not
              # Unicode characters :-( RFC 9110, section 5.5. We can
              # have end-of-lines, too, if they are indented, but we
              # don't handle that case.
  header_line = header_name |> ignore(string(":")) |> ignore(repeat(string(" ")))
                |> concat(header_value) |> ignore(eol) # We can have
             # zero or more spaces after the semi-colon ('repeat' can
             # be zero times), but not before. RFC 9112 section 5 and
             # 9110, section 5.6.3.
  headers = header_line |> wrap() |> repeat() |> wrap() |> ignore(eol) # The empty
             # line at the end of headers is mandatory (RFC 9112, section 2.1).
  body = choice([ascii_string([], min: 0), # An empty body is allowed.
		 empty() |> replace("")]) # No body is also allowed
      # but we have to return a token, here the empty string.
  
  # We insist on one space only. The RFC 9112 section 3 recommends it.
  defparsec :request, method |> ignore(string(" "))
                      |> concat(path) |> ignore(string(" "))
                      |> concat(version) |> ignore(eol) |> concat(headers) |> optional(body) |> eos

  # This is not really satisfactory since a compliant HTTP parser
  # would need to be incremental because the request may not be sent in
  # a single TCP segment. We need a parser where you can stream data
  # but it does not seem possible with NimbleParsec. You can see the
  # problem when using telnet as a client, since it sends every line
  # in a TCP segment, the parser may stop too early.

  # Also, properly reading the body requires to know the header
  # (Content-Length or Transfer-Encoding) so you cannot really have a
  # grammar rule with the body. But our goal is not to be perfect but
  # to illustrate HTTP.
  
end

